#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd

from test.common import initData, TESTFLAGGER, TESTNODATA
from saqc.core.core import run
from saqc.core.config import Fields as F

from test.common import initData, TESTFLAGGER, TESTNODATA, initMetaDict

from saqc.core.evaluator import (
    DslTransformer,
    initLocalEnv,
    parseExpression,
    evalExpression,
    compileTree,
    evalCode,
)


def _evalDslExpression(expr, data, field, flagger, nodata=np.nan):
    env = initLocalEnv(data, field, flagger, nodata)
    tree = parseExpression(expr)
    transformed_tree = DslTransformer(env).visit(tree)
    code = compileTree(transformed_tree)
    return evalCode(code, local_env=env)


@pytest.fixture
def data():
    return initData()


# @pytest.mark.parametrize("flagger", TESTFLAGGER)
# def test_flagPropagation(data, flagger):
#     var1, var2, *_ = data.columns
#     this = var1

#     flagger = flagger.initFlags(data).setFlags(var2, iloc=slice(None, None, 5))

#     var2_flags = flagger.isFlagged(var2)
#     var2_data = data[var2].mask(var2_flags)
#     data, flagger_result = evalExpression(
#         "flagGeneric(func=var2 < mean(var2))", data, this, flagger, np.nan
#     )

#     expected = var2_flags | (var2_data < var2_data.mean())
#     result = flagger_result.isFlagged(this)
#     assert (result == expected).all()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_missingIdentifier(data, flagger):

    flagger = flagger.initFlags(data)
    tests = [
        "flagGeneric(func=fff(var2) < 5)",
        "flagGeneric(func=var3 != NODATA)"
    ]
    for expr in tests:
        with pytest.raises(NameError):
            evalExpression(expr, data, data.columns[0], flagger, np.nan)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_comparisonOperators(data, flagger):
    flagger = flagger.initFlags(data)
    var1, var2, *_ = data.columns
    this = var1

    tests = [
        ("this > 100", data[this] > 100),
        (f"10 >= {var2}", 10 >= data[var2]),
        (f"{var2} < 100", data[var2] < 100),
        (f"this <= {var2}", data[this] <= data[var2]),
        (f"{var1} == {var2}", data[this] == data[var2]),
        (f"{var1} != {var2}", data[this] != data[var2]),
    ]

    # check within the usually enclosing scope
    for expr, mask in tests:
        _, result_flagger = evalExpression(
            f"flagGeneric(func={expr})", data, this, flagger, np.nan
        )
        expected_flagger = flagger.setFlags(this, loc=mask, test="generic")
        assert np.all(result_flagger.isFlagged() == expected_flagger.isFlagged())


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_arithmeticOperators(data, flagger):
    flagger = flagger.initFlags(data)
    var1, *_ = data.columns
    this = data[var1]

    tests = [
        ("this + 100 > 110", this + 100 > 110),
        ("this - 100 > 0", this - 100 > 0),
        ("this * 100 > 200", this * 100 > 200),
        ("this / 100 > .1", this / 100 > .1),
        ("this % 2 == 1", this % 2 == 1),
        ("this ** 2 == 0", this ** 2 == 0),
    ]

    # check within the usually enclosing scope
    for expr, mask in tests:
        _, result_flagger = evalExpression(
            f"flagGeneric(func={expr})", data, var1, flagger, np.nan
        )
        expected_flagger = flagger.setFlags(var1, loc=mask, test="generic")
        assert np.all(result_flagger.isFlagged() == expected_flagger.isFlagged())


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_nonReduncingBuiltins(data, flagger):
    flagger = flagger.initFlags(data)
    var1, *_ = data.columns
    this = var1

    tests = [
        ("abs(this)", np.abs(data[this])),
    ]

    for expr, expected in tests:
        result = _evalDslExpression(expr, data, this, flagger)
        assert (result == expected).all()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("nodata", TESTNODATA)
def test_reduncingBuiltins(data, flagger, nodata):
    data.loc[::4] = nodata
    flagger = flagger.initFlags(data)
    var1, var2, *_ = data.columns
    this = var1

    tests = [
        ("min(this)", np.min(data[this])),
        (f"max({var1})", np.max(data[var1])),
        (f"sum({var2})", np.sum(data[var2])),
        ("mean(this)", np.mean(data[this])),
        (f"std({var1})", np.std(data[var1])),
        (f"len({var2})", len(data[var2])),
    ]

    for expr, expected in tests:
        result = _evalDslExpression(expr, data, this, flagger, nodata)
        assert result == expected


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("nodata", TESTNODATA)
def test_ismissing(data, flagger, nodata):

    data.iloc[: len(data) // 2, 0] = np.nan
    data.iloc[(len(data) // 2) + 1 :, 0] = -9999
    var1, *_ = data.columns

    flagger = flagger.initFlags(data)

    tests = [
        (f"ismissing({var1})", lambda data: (pd.isnull(data) | (data == nodata)).all()),
        (
            f"~ismissing({var1})",
            lambda data: (pd.notnull(data) & (data != nodata)).all(),
        ),
    ]

    for expr, checkFunc in tests:
        idx = _evalDslExpression(expr, data, var1, flagger, nodata)
        assert checkFunc(data.loc[idx, var1])


@pytest.mark.parametrize("flagger", TESTFLAGGER)
@pytest.mark.parametrize("nodata", TESTNODATA)
def test_bitOps(data, flagger, nodata):
    var1, var2, *_ = data.columns
    this = var1

    flagger = flagger.initFlags(data)

    tests = [
        (f"flagGeneric(func=~(this > mean(this)))", ~(data[this] > np.nanmean(data[this]))),
        (
            f"flagGeneric(func=(this <= 0) | (0 < {var1}))",
            (data[this] <= 0) | (0 < data[var1]),
        ),
        (
            f"flagGeneric(func=({var2} >= 0) & (0 > this))",
            (data[var2] >= 0) & (0 > data[this]),
        ),
    ]

    for expr, expected in tests:
        _, flagger_result = evalExpression(expr, data, this, flagger, nodata)
        assert (flagger_result.isFlagged(this) == expected).all()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_isflagged(data, flagger):

    flagger = flagger.initFlags(data)
    var1, var2, *_ = data.columns

    flagger = flagger.setFlags(var1, iloc=slice(None, None, 2))
    flagger = flagger.setFlags(var2, iloc=slice(None, None, 2))

    idx = _evalDslExpression(f"isflagged({var1})", data, var2, flagger)

    flagged = flagger.isFlagged(var1)
    assert (flagged == idx).all


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_invertIsFlagged(data, flagger):

    flagger = flagger.initFlags(data)
    var1, var2, *_ = data.columns

    flagger = flagger.setFlags(var2, iloc=slice(None, None, 2))

    tests = [
        (f"~isflagged({var2})", ~flagger.isFlagged(var2)),
        (f"~({var2}>999) & (~isflagged({var2}))", ~(data[var2] > 999) & (~flagger.isFlagged(var2)))
    ]

    for expr, flags_expected in tests:
        _, flagger_result = evalExpression(
            f"flagGeneric(func={expr})", data, var1, flagger, np.nan
        )
        flags_result = flagger_result.isFlagged(var1)
        assert np.all(flags_result == flags_expected)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_isflaggedArgument(data, flagger):

    var1, var2, *_ = data.columns

    flagger = flagger.initFlags(data).setFlags(
        var1, iloc=slice(None, None, 2), flag=flagger.BAD
    )

    tests = [
        (_evalDslExpression(f"isflagged({var1}, BAD)", data, var2, flagger),
         flagger.isFlagged(var1, flag=flagger.BAD)
        ),
        (_evalDslExpression(f"isflagged({var1}, UNFLAGGED, '==')", data, var2, flagger),
         flagger.isFlagged(var1, flag=flagger.UNFLAGGED, comparator="==")),
    ]

    for result, expected in tests:
        assert np.all(result == expected)




#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd

from scipy.signal import savgol_filter

from saqc.funcs.register import register
from saqc.lib.tools import retrieveTrustworthyOriginal


@register("breaks_spektrumBased")
def flagBreaksSpektrumBased(
    data,
    field,
    flagger,
    thresh_rel=0.1,
    thresh_abs=0.01,
    first_der_factor=10,
    first_der_window="12h",
    scnd_der_ratio_range=0.05,
    scnd_der_ratio_thresh=10,
    smooth=True,
    smooth_window=None,
    smooth_poly_deg=2,
    **kwargs
):

    """ This Function is an generalization of the Spectrum based break flagging mechanism as presented in:

    Dorigo,W,.... Global Automated Quality Control of In Situ Soil Moisture Data from the international
    Soil Moisture Network. 2013. Vadoze Zone J. doi:10.2136/vzj2012.0097.

    The function flags breaks (jumps/drops) in input measurement series by evaluating its derivatives.
    A measurement y_t is flagged a, break, if:

    (1) y_t is changing relatively to its preceeding value by at least (100*rel_change_rate_min) percent
    (2) y_(t-1) is difffering from its preceeding value, by a margin of at least "thresh_abs"
    (3) Absolute first derivative |(y_t)'| has to be at least "first_der_factor" times as big as the arithmetic middle
        over all the first derivative values within a 2 times "first_der_window_size" hours window, centered at t.
    (4) The ratio of the second derivatives at t and t+1 has to be "aproximately" 1.
        ([1-scnd__der_ration_margin_1, 1+scnd_ratio_margin_1])
    (5) The ratio of the second derivatives at t+1 and t+2 has to be larger than scnd_der_ratio_margin_2

    NOTE 1: As no reliable statement about the plausibility of the meassurements before and after the jump is possible,
    only the jump itself is flagged. For flagging constant values following upon a jump, use a flagConstants test.

    NOTE 2: All derivatives in the reference publication are obtained by applying a Savitzky-Golay filter to the data
    before differentiating. However, i was not able to reproduce satisfaction of all the conditions for synthetically
    constructed breaks.
    Especially condition [4] and [5]! This is because smoothing distributes the harshness of the break over the
    smoothing window. Since just taking the differences as derivatives did work well for my empirical data set,
    the parameter "smooth" defaults to "raw". That means, that derivatives will be obtained by just using the
    differences series.
    You are free of course, to change this parameter to "savgol" and play around with the associated filter options.
    (see parameter description below)




       :param data:                        The pandas dataframe holding the data-to-be flagged.
                                           Data must be indexed by a datetime series and be harmonized onto a
                                           time raster with seconds precision (skips allowed).
       :param flags:                       A dataframe holding the flags/flag-entries associated with "data".
       :param field:                       Fieldname of the Soil moisture measurements field in data.
       :param flagger:                     A flagger - object. (saqc.flagger.X)
       :param smooth:                      Bool. Method for obtaining dataseries' derivatives.
                                           False: Just take series step differences (default)
                                           True: Smooth data with a Savitzky Golay Filter before differentiating.
       :param smooth_window:               Offset string. Size of the filter window, used to calculate the derivatives.
                                           (relevant only, if: smooth is True)
       :param smooth_poly_deg:             Integer. Polynomial order, used for smoothing with savitzk golay filter.
                                           (relevant only, if: smooth_func='savgol')
       :param thresh_rel                   Float in [0,1]. See (1) of function descritpion above to learn more
       :param thresh_abs                   Float > 0. See (2) of function descritpion above to learn more.
       :param first_der_factor             Float > 0. See (3) of function descritpion above to learn more.
       :param first_der_window_range        Offset_String. See (3) of function description to learn more.
       :param scnd_der_ratio_margin_1      Float in [0,1]. See (4) of function descritpion above to learn more.
       :param scnd_der_ratio_margin_2      Float in [0,1]. See (5) of function descritpion above to learn more.
    """

    # retrieve data series input at its original sampling rate
    dataseries, data_rate = retrieveTrustworthyOriginal(data, field, flagger)

    if smooth_window is None:
        smooth_window = 3 * pd.Timedelta(data_rate)
    else:
        smooth_window = pd.Timedelta(smooth_window)

    # relative - change - break criteria testing:
    abs_change = np.abs(dataseries.shift(+1) - dataseries)
    breaks = (abs_change > thresh_abs) & (abs_change / dataseries > thresh_rel)
    breaks = breaks[breaks == True]

    # First derivative criterion
    smoothing_periods = int(np.ceil((smooth_window.seconds / data_rate.n)))
    if smoothing_periods % 2 == 0:
        smoothing_periods += 1

    for brake in breaks.index:
        # slice out slice-to-be-filtered (with some safety extension of 12 times the data rate)
        slice_start = brake - pd.Timedelta(first_der_window) - smoothing_periods * pd.Timedelta(data_rate)
        slice_end = brake + pd.Timedelta(first_der_window) + smoothing_periods * pd.Timedelta(data_rate)
        data_slice = dataseries[slice_start:slice_end]

        # obtain first derivative:
        if smooth is True:
            first_deri_series = pd.Series(
                data=savgol_filter(data_slice, window_length=smoothing_periods, polyorder=smooth_poly_deg, deriv=1,),
                index=data_slice.index,
            )
        else:
            first_deri_series = data_slice.diff()

        # condition constructing and testing:
        test_slice = first_deri_series[brake - pd.Timedelta(first_der_window) : brake + pd.Timedelta(first_der_window)]

        test_sum = abs((test_slice.sum() * first_der_factor) / test_slice.size)

        if abs(first_deri_series[brake]) > test_sum:
            # second derivative criterion:
            slice_start = brake - 12 * pd.Timedelta(data_rate)
            slice_end = brake + 12 * pd.Timedelta(data_rate)
            data_slice = data_slice[slice_start:slice_end]

            # obtain second derivative:
            if smooth is True:
                second_deri_series = pd.Series(
                    data=savgol_filter(
                        data_slice, window_length=smoothing_periods, polyorder=smooth_poly_deg, deriv=2,
                    ),
                    index=data_slice.index,
                )
            else:
                second_deri_series = data_slice.diff().diff()

            # criterion evaluation:
            first_second = (
                (1 - scnd_der_ratio_range)
                < abs((second_deri_series.shift(+1)[brake] / second_deri_series[brake]))
                < 1 + scnd_der_ratio_range
            )

            second_second = abs(second_deri_series[brake] / second_deri_series.shift(-1)[brake]) > scnd_der_ratio_thresh

            if (~first_second) | (~second_second):
                breaks[brake] = False

        else:
            breaks[brake] = False

    breaks = breaks[breaks == True]

    flagger = flagger.setFlags(field, breaks.index, **kwargs)

    return data, flagger



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from saqc.flagger.baseflagger import BaseFlagger
from saqc.flagger.categoricalflagger import CategoricalFlagger
from saqc.flagger.simpleflagger import SimpleFlagger
from saqc.flagger.dmpflagger import DmpFlagger
from saqc.flagger.continuousflagger import ContinuousFlagger



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np

from saqc.funcs.constants_detection import flagConstantVarianceBased, flagConstant
from saqc.funcs.soil_moisture_tests import flagSoilMoistureConstant

from test.common import TESTFLAGGER, initData


@pytest.fixture
def data():
    constants_data = initData(
        1, start_date="2011-01-01 00:00:00", end_date="2011-01-01 03:00:00", freq="5min"
    )
    constants_data.iloc[5:25] = 0
    return constants_data


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagConstants(data, flagger):
    idx = np.array([5, 6, 7, 8, 9, 10, 18, 19, 20, 21])
    data.iloc[idx] = 200
    expected = np.arange(5, 22)
    field, *_ = data.columns
    flagger = flagger.initFlags(data)
    data, flagger_result = flagConstant(
        data, field, flagger, window="15Min", thresh=.1,
    )
    flags = flagger_result.getFlags(field)
    assert np.all(flags[expected] == flagger.BAD)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagConstantsVarianceBased(data, flagger):
    data.iloc[5:25] = 200
    expected = np.arange(5, 25)
    field, *_ = data.columns
    flagger = flagger.initFlags(data)
    data, flagger_result1 = flagConstantVarianceBased(
        data, field, flagger, window="1h"
    )

    flag_result1 = flagger_result1.getFlags(field)
    test_sum = (flag_result1[expected] == flagger.BAD).sum()
    assert test_sum == len(expected)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import ast
from saqc.core.config import Params
from typing import Dict, Any


class DslTransformer(ast.NodeTransformer):
    def __init__(self, environment: Dict[str, Any]):
        self.environment = environment
        self.arguments = set()

    def visit_Call(self, node):
        return ast.Call(func=node.func, args=[self.visit(arg) for arg in node.args], keywords=[])

    def visit_Name(self, node):
        name = node.id

        if name == "this":
            name = self.environment["field"]

        if name in self.environment["variables"]:
            value = ast.Constant(value=name)
            node = ast.Subscript(
                value=ast.Name(id="data", ctx=ast.Load()), slice=ast.Index(value=value), ctx=ast.Load(),
            )

        self.arguments.add(name)
        return node


class ConfigTransformer(ast.NodeTransformer):
    def __init__(self, environment):
        self.environment = environment
        self.func_name = None

    def visit_Call(self, node):
        self.func_name = node.func.id

        new_args = [
            ast.Name(id="data", ctx=ast.Load()),
            ast.Name(id="field", ctx=ast.Load()),
            ast.Name(id="flagger", ctx=ast.Load()),
        ]
        node = ast.Call(func=node.func, args=new_args + node.args, keywords=node.keywords)

        return self.generic_visit(node)

    def visit_keyword(self, node):
        key, value = node.arg, node.value

        if self.func_name == Params.FLAG_GENERIC and key == Params.FUNC:
            dsl_transformer = DslTransformer(self.environment)
            value = dsl_transformer.visit(value)
            dsl_func = ast.keyword(arg=key, value=value)
            # NOTE:
            # Inject the additional `func_arguments` argument `flagGeneric`
            # expects, to keep track of all the touched variables. We
            # need this to propagate the flags from the independent variables
            args = ast.keyword(
                arg=Params.GENERIC_ARGS,
                value=ast.List(elts=[ast.Str(s=v) for v in dsl_transformer.arguments], ctx=ast.Load(),),
            )
            return [dsl_func, args]

        return self.generic_visit(node)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np

from saqc.funcs import register
from saqc.core.evaluator import (
    compileTree,
    parseExpression,
    initLocalEnv,
    DslTransformer,
    ConfigChecker,
    ConfigTransformer,
)

from test.common import TESTFLAGGER, initData


def compileExpression(expr, flagger, nodata=np.nan):
    data = initData()
    field = data.columns[0]
    tree = parseExpression(expr)
    env = initLocalEnv(data, field, flagger.initFlags(data), nodata)
    ConfigChecker(env, flagger.signature).visit(tree)
    transformed_tree = ConfigTransformer(env).visit(tree)
    code = compileTree(transformed_tree)
    return code


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_syntaxError(flagger):
    exprs = [
        "range(x=5",
        "rangex=5)",
        "range[x=5]" "range{x=5}" "int->float(x=4)" "int*float(x=4)",
    ]

    for expr in exprs:
        with pytest.raises(SyntaxError):
            compileExpression(expr, flagger)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_typeError(flagger):

    exprs = [
        # "func",
        "dummy(kwarg=[1, 2, 3])",
        "dummy(x=5)",
        "dummy(dummy())",
        "dummy(kwarg=dummy(this))",
    ]

    for expr in exprs:
        with pytest.raises(TypeError):
            compileExpression(expr, flagger)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_supportedArguments(flagger):

    @register("func")
    def dummyRegisterFunc(data, field, flagger, kwarg, **kwargs):
        return data, flagger

    exprs = [
        "func(kwarg='str')",
        "func(kwarg=5)",
        "func(kwarg=5.5)",
        "func(kwarg=-5)",
        "func(kwarg=True)",
        "func(kwarg=func())",
    ]
    for expr in exprs:
        compileExpression(expr, flagger)





#! /usr/bin/env python
# -*- coding: utf-8 -*-

# imports needed to make the functions register themself
from .register import register
from .functions import *
from .break_detection import *
from .constants_detection import *
from .soil_moisture_tests import *
from .spike_detection import *
from .harm_functions import *
from .machine_learning import *



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import ast

from saqc.funcs.register import FUNC_MAP
from saqc.core.config import Params


class DslChecker(ast.NodeVisitor):

    SUPPORTED = (
        ast.Expression,
        ast.UnaryOp,
        ast.BinOp,
        ast.BitOr,
        ast.BitAnd,
        ast.Num,
        ast.Compare,
        ast.Add,
        ast.Sub,
        ast.Mult,
        ast.Div,
        ast.Pow,
        ast.Mod,
        ast.USub,
        ast.Eq,
        ast.NotEq,
        ast.Gt,
        ast.Lt,
        ast.GtE,
        ast.LtE,
        ast.Invert,
        ast.Name,
        ast.Load,
        ast.Call,
    )

    def __init__(self, environment):
        self.environment = environment

    def visit_Call(self, node):
        func_name = node.func.id
        if func_name not in self.environment:
            raise NameError(f"unspported function: '{func_name}'")
        self.generic_visit(node)

    def visit_Name(self, node):
        name = node.id
        if name != "this" and name not in self.environment and name not in self.environment["variables"]:
            raise NameError(f"unknown variable: '{name}'")
        self.generic_visit(node)

    def generic_visit(self, node):
        if not isinstance(node, self.SUPPORTED):
            raise TypeError(f"invalid expression: '{node}'")
        return super().generic_visit(node)


class ConfigChecker(ast.NodeVisitor):

    SUPPORTED_NODES = (
        ast.Call,
        ast.Num,
        ast.Str,
        ast.keyword,
        ast.NameConstant,
        ast.UnaryOp,
        ast.Name,
        ast.Load,
        ast.Expression,
        ast.Subscript,
        ast.Index,
        ast.USub,
    )

    SUPPORTED_ARGUMENTS = (
        ast.Str,
        ast.Num,
        ast.NameConstant,
        ast.Call,
        ast.UnaryOp,
        ast.USub,
        ast.Name,
    )

    def __init__(self, environment, pass_parameter):
        self.pass_parameter = pass_parameter
        self.environment = environment
        self.func_name = None

    def visit_Call(self, node):

        func_name = node.func.id
        if func_name not in FUNC_MAP:
            raise NameError(f"unknown test function: '{func_name}'")
        if node.args:
            raise TypeError("only keyword arguments are supported")
        self.func_name = func_name
        return self.generic_visit(node)

    def visit_keyword(self, node):
        key, value = node.arg, node.value
        if self.func_name == Params.FLAG_GENERIC and key == Params.FUNC:
            DslChecker(self.environment).visit(value)
            return

        if key not in FUNC_MAP[self.func_name].signature + self.pass_parameter:
            raise TypeError(f"unknown function parameter '{node.arg}'")

        if not isinstance(value, self.SUPPORTED_ARGUMENTS):
            raise TypeError(f"invalid argument type '{type(value)}'")

        if isinstance(value, ast.Name) and value.id not in self.environment:
            raise NameError(f"unknown variable: {value.id}")

        return self.generic_visit(node)

    def generic_visit(self, node):
        if not isinstance(node, self.SUPPORTED_NODES):
            raise TypeError(f"invalid node: '{node}'")
        return super().generic_visit(node)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import numpy as np
import pandas as pd
from scipy.signal import savgol_filter
from scipy.stats import zscore
from saqc.funcs.register import register
import numpy.polynomial.polynomial as poly

from saqc.lib.tools import (
    inferFrequency,
    retrieveTrustworthyOriginal,
    offset2seconds,
    slidingWindowIndices,
)


@register("spikes_slidingZscore")
def flagSpikes_slidingZscore(
    data, field, flagger, window, offset, count=1, polydeg=1, z=3.5, method="modZ", **kwargs,
):
    """ A outlier detection in a sliding window. The method for detection can be a simple Z-score or the more robust
    modified Z-score, as introduced here [1].

    The steps are:
    1.  a window of size `window` is cut from the data
    2.  the data is fit by a polynomial of the given degree `polydeg`
    3.  the outlier `method` detect potential outlier
    4.  the window is continued by `offset` to the next data-slot.
    5.  processing continue at 1. until end of data.
    6.  all potential outlier, that are detected `count`-many times, are promoted to real outlier and flagged by the `flagger`

    :param data:        pandas dataframe. holding the data
    :param field:       fieldname in `data`, which holds the relevant infos
    :param flagger:     flagger.
    :param window:      int or time-offset string (see [2]). The size of the window the outlier detection is run in. default: 1h
    :param offset:      int or time-offset string (see [2]). Stepsize the window is set further. default: 1h
    :param method:      str. `modZ`  or `zscore`. see [1] at section `Z-Scores and Modified Z-Scores`
    :param count:       int. this many times, a datapoint needs to be detected in different windows, to be finally
                        flagged as outlier
    :param polydeg:     The degree for the polynomial fit, to calculate the residuum
    :param z:           float. the value the (mod.) Z-score is tested against. Defaulting to 3.5 (Recommendation of [1])

    Links:
    [1] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
    [2] https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
    """

    use_offset = False
    dx_s = offset
    winsz_s = window
    # check param consistency
    if isinstance(window, str) or isinstance(offset, str):
        if isinstance(window, str) and isinstance(offset, str):
            use_offset = True
            dx_s = offset2seconds(offset)
            winsz_s = offset2seconds(window)
        else:
            raise TypeError(
                f"`window` and `offset` must both be an offset or both be numeric, {window} and {offset} was passed"
            )

    # check params
    if polydeg < 0:
        raise ValueError("polydeg must be positive")
    if z < 0:
        raise ValueError("z must be positive")
    if count <= 0:
        raise ValueError("count must be positive and not zero")

    if dx_s >= winsz_s and count == 1:
        pass
    elif dx_s >= winsz_s and count > 1:
        ValueError("If stepsize `offset` is bigger that the window-size, every value is seen just once, so use count=1")
    elif count > winsz_s // dx_s:
        raise ValueError(
            f"Adjust `offset`, `stepsize` or `window`. A single data point is "
            f"seen `floor(window / offset) = {winsz_s // dx_s}` times, but count is set to {count}"
        )

    # prepare the method
    method = method.lower()
    if method == "modz":

        def _calc(residual):
            diff = np.abs(residual - np.median(residual))
            mad = np.median(diff)
            return (mad > 0) & (0.6745 * diff > z * mad)

    elif method == "zscore":

        def _calc(residual):
            score = zscore(residual, ddof=1)
            return np.abs(score) > z

    else:
        raise NotImplementedError
    method = _calc

    # prepare data, work on numpy arrays for the fulfilling pleasure of performance
    d = data[field].dropna()
    all_indices = np.arange(len(d.index))
    x = (d.index - d.index[0]).total_seconds().values
    y = d.values
    counters = np.full(len(d.index), count)

    if use_offset:
        _loopfun = slidingWindowIndices
    else:

        def _loopfun(arr, wsz, step):
            for i in range(0, len(arr) - wsz + 1, step):
                yield i, i + wsz

    for start, end in _loopfun(d.index, window, offset):
        # mask points that have been already discarded
        mask = counters[start:end] > 0
        indices = all_indices[all_indices[start:end][mask]]
        xchunk = x[indices]
        ychunk = y[indices]

        if xchunk.size == 0:
            continue

        # get residual
        coef = poly.polyfit(xchunk, ychunk, polydeg)
        model = poly.polyval(xchunk, coef)
        residual = ychunk - model

        score = method(residual)

        # count`em in
        goneMad = score.nonzero()[0]
        counters[indices[goneMad]] -= 1

    outlier = np.where(counters <= 0)[0]
    loc = d[outlier].index
    flagger = flagger.setFlags(field, loc=loc, **kwargs)
    return data, flagger


@register("spikes_simpleMad")
def flagSpikes_simpleMad(data, field, flagger, window, z=3.5, **kwargs):
    """ The function represents an implementation of the modyfied Z-score outlier detection method, as introduced here:

    [1] https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm

    The test needs the input data to be harmonized to an equidustant time stamp series (=have frequencie))

    :param data:        The pandas dataframe holding the data-to-be flagged.
                        Data must be indexed by a datetime series and be harmonized onto a
                        time raster with seconds precision.
    :param field:       Fieldname of the Soil moisture measurements field in data.
    :param flagger:     A flagger - object. (saqc.flagger.X)
    :param winsz:      Offset String. Denoting the windows size that that th "Z-scored" values have to lie in.
    :param z:           Float. The value the Z-score is tested against. Defaulting to 3.5 (Recommendation of [1])
    """
    d = data[field].copy().mask(flagger.isFlagged(field))
    median = d.rolling(window=window, closed="both").median()
    diff = (d - median).abs()
    mad = diff.rolling(window=window, closed="both").median()
    mask = (mad > 0) & (0.6745 * diff > z * mad)
    # NOTE:
    # In pandas <= 0.25.3, the window size is not fixed if the
    # window-argument to rolling is a frequency. That implies,
    # that during the first iterations the window has a size of
    # 1, 2, 3, ... until it eventually covers the disered time
    # span. For stuff the calculation of median, that is rather
    # unfortunate, as the size of calculation base might differ
    # heavily. So don't flag something until, the window reaches
    # its target size
    if not isinstance(window, int):
        index = mask.index
        mask.loc[index < index[0] + pd.to_timedelta(window)] = False

    flagger = flagger.setFlags(field, mask, **kwargs)
    return data, flagger


@register("spikes_basic")
def flagSpikes_basic(data, field, flagger, thresh=7, tolerance=0, window="15min", **kwargs):
    """
    A basic outlier test that is designed to work for harmonized and not harmonized data.

    Values x(n), x(n+1), .... , x(n+k) of a timeseries x are considered spikes, if

    (1) |x(n-1) - x(n + s)| > "thresh", for all s in [0,1,2,...,k]

    (2) |x(n-1) - x(n+k+1)| < tol

    (3) |x(n-1).index - x(n+k+1).index| < length

    Note, that this definition of a "spike" not only includes one-value outliers, but also plateau-ish value courses.

    The implementation is a time-window based version of an outlier test from the UFZ Python library,
    that can be found here:

    https://git.ufz.de/chs/python/blob/master/ufz/level1/spike.py


    :param data:    Pandas-like. The pandas dataframe holding the data-to-be flagged.
    :param field:   String. Fieldname of the data column to be tested.
    :param flagger: saqc.flagger. A flagger - object.
    :param thresh:  Float. The lower bound for a value jump, to be considered as initialising a spike.
                    (see condition (1) in function description).
    :param tolerance: Float. Tolerance value.  (see condition (2) in function description)
    :param window_size:  Offset String. The time span in wich the values of a spikey course have to return to the normal
                    value course (see condition (3) in function description).
    :return:
    """

    dataseries = data[field].dropna()
    # get all the entries preceding a significant jump
    pre_jumps = dataseries.diff(periods=-1).abs() > thresh
    pre_jumps = pre_jumps[pre_jumps]
    if pre_jumps.empty:
        return data, flagger
    # get all the entries preceeding a significant jump and its successors within "length" range
    to_roll = pre_jumps.reindex(dataseries.index, method="ffill", tolerance=window, fill_value=False).dropna()

    # define spike testing function to roll with:
    def spike_tester(chunk, pre_jumps_index, thresh, tol):
        if not chunk.index[-1] in pre_jumps_index:
            return 0
        else:
            # signum change!!!
            chunk_stair = (abs(chunk - chunk[-1]) < thresh)[::-1].cumsum()
            first_return = chunk_stair[(chunk_stair == 2)]
            if first_return.sum() == 0:
                return 0
            if abs(chunk[first_return.index[0]] - chunk[-1]) < tol:
                return (chunk_stair == 1).sum() - 1
            else:
                return 0

    # since .rolling does neither support windows, defined by left starting points, nor rolling over monotonically
    # decreasing indices, we have to trick the method by inverting the timeseries and transforming the resulting index
    # to pseudo-increase.
    to_roll = dataseries[to_roll]
    original_index = to_roll.index
    to_roll = to_roll[::-1]
    pre_jump_reversed_index = to_roll.index[0] - pre_jumps.index
    to_roll.index = to_roll.index[0] - to_roll.index

    # now lets roll:
    to_roll = (
        to_roll.rolling(window, closed="both")
        .apply(spike_tester, args=(pre_jump_reversed_index, thresh, tolerance), raw=False)
        .astype(int)
    )
    # reconstruct original index and sequence
    to_roll = to_roll[::-1]
    to_roll.index = original_index
    to_write = to_roll[to_roll != 0]
    to_flag = pd.Index([])
    # here comes a loop...):
    for row in to_write.iteritems():
        loc = to_roll.index.get_loc(row[0])
        to_flag = to_flag.append(to_roll.iloc[loc + 1 : loc + row[1] + 1].index)

    to_flag = to_flag.drop_duplicates(keep="first")
    flagger = flagger.setFlags(field, to_flag, **kwargs)
    return data, flagger


@register("spikes_spektrumBased")
def flagSpikes_spektrumBased(
    data,
    field,
    flagger,
    raise_factor=0.15,
    deriv_factor=0.2,
    noise_func="CoVar",
    noise_window="12h",
    noise_thresh=1,
    smooth_window=None,
    smooth_poly_deg=2,
    **kwargs,
):
    """
    This Function is a generalization of the Spectrum based Spike flagging mechanism as presented in:

    Dorigo,W,.... Global Automated Quality Control of In Situ Soil Moisture Data from the international
    Soil Moisture Network. 2013. Vadoze Zone J. doi:10.2136/vzj2012.0097.

    Function detects and flags spikes in input data series by evaluating its derivatives and applying some
    conditions to it. A datapoint is considered a spike, if:

    (1) the quotient to its preceeding datapoint exceeds a certain bound
    (controlled by param "raise_factor")
    (2) the quotient of the datas second derivate at the preceeding and subsequent timestamps is close enough to 1.
    (controlled by param "deriv_factor")
    (3) the surrounding data is not too noisy. (Coefficient of Variation[+/- noise_window] < 1)
    (controlled by param "noise_thresh")

    Some things you should be conscious about when applying this test:

       NOTE1: You should run less complex tests, especially range-tests, or absolute spike tests previously to this one,
       since the spike check for any potential, unflagged spike, is relatively costly
       (1 x smoothing + 2 x deviating + 2 x condition application).

       NOTE2: Due to inconsistency in the paper that provided the concept of this test [paper:], its not really clear
       weather to use the coefficient of variance or the relative variance for noise testing.
       Since the relative variance was explicitly denoted in the formulas, the function defaults to relative variance,
       but can be switched to coefficient of variance, by assignment to parameter "noise statistic".


       :param data:                 The pandas dataframe holding the data-to-be flagged.
                                    Data must be indexed by a datetime series and be harmonized onto a
                                    time raster with seconds precision.
       :param field:                Fieldname of the Soil moisture measurements field in data.
       :param flagger:              A flagger - object. (saqc.flagger.X)
       :param smooth_window:        Offset string. Size of the filter window, used to calculate the derivatives.
                                    (relevant only, if: diff_method='savgol')
       :param smooth_poly_deg:      Integer. Polynomial order, used for smoothing with savitzk golay filter.
                                    (relevant only, if: diff_method='savgol')
       :param raise_factor:         A float, determinating the bound, the quotient of two consecutive values
                                    has to exceed, to be regarded as potentially spike. A value of 0.x will
                                    trigger the spike test for value y_t, if:
                                    (y_t)/(y_t-1) > 1 + x or:
                                    (y_t)/(y_t-1) < 1 - x.
       :param deriv_factor:         A float, determining the interval, the quotient of the datas second derivate
                                    around a potential spike has to be part of, to trigger spike flagging for
                                    this value. A datapoint y_t will pass this spike condition if,
                                    for deriv_factor = 0.x, and the second derivative y'' of y, the condition:
                                    1 - x < abs((y''_t-1)/(y''_t+1)) < 1 + x
                                    holds
       :param noise_thresh:         A float, determining the bound, the data noisy-ness around a potential spike
                                    must not exceed, in order to guarantee a justifyed judgement:
                                    Therefor the coefficient selected by parameter noise_func (COVA),
                                    of all values within t +/- param "noise_window",
                                    but excluding the point y_t itself, is evaluated and tested
                                    for: COVA < noise_thresh.
       :param noise_window:         Offset string, determining the size of the window, the coefficient of
                                    variation is calculated of, to determine data noisy-ness around a potential
                                    spike.
                                    The potential spike y_t will be centered in a window of expansion:
                                    [y_t - noise_window_size, y_t + noise_window_size].
       :param noise_func:           String. Determines, wheather to use
                                    "relative variance" or "coefficient of variation" to check against the noise
                                    barrier.
                                    'CoVar' -> "Coefficient of variation"
                                    'rVar'  -> "relative Variance"
    """

    dataseries, data_rate = retrieveTrustworthyOriginal(data, field, flagger)
    noise_func_map = {"covar": pd.Series.var, "rvar": pd.Series.std}
    noise_func = noise_func_map[noise_func.lower()]

    if smooth_window is None:
        smooth_window = 3 * pd.Timedelta(data_rate)
    else:
        smooth_window = pd.Timedelta(smooth_window)

    quotient_series = dataseries / dataseries.shift(+1)
    spikes = (quotient_series > (1 + raise_factor)) | (quotient_series < (1 - raise_factor))
    spikes = spikes[spikes == True]

    # loop through spikes: (loop may sound ugly - but since the number of spikes is supposed to not exceed the
    # thousands for year data - a loop going through all the spikes instances is much faster than
    # a rolling window, rolling all through a stacked year dataframe )

    # calculate some values, repeatedly needed in the course of the loop:

    filter_window_seconds = smooth_window.seconds
    smoothing_periods = int(np.ceil((filter_window_seconds / data_rate.n)))
    lower_dev_bound = 1 - deriv_factor
    upper_dev_bound = 1 + deriv_factor

    if smoothing_periods % 2 == 0:
        smoothing_periods += 1

    for spike in spikes.index:
        start_slice = spike - smooth_window
        end_slice = spike + smooth_window

        scnd_derivate = savgol_filter(
            dataseries[start_slice:end_slice], window_length=smoothing_periods, polyorder=smooth_poly_deg, deriv=2,
        )

        length = scnd_derivate.size
        test_ratio_1 = np.abs(scnd_derivate[int(((length + 1) / 2) - 2)] / scnd_derivate[int(((length + 1) / 2))])

        if lower_dev_bound < test_ratio_1 < upper_dev_bound:
            # apply noise condition:
            start_slice = spike - pd.Timedelta(noise_window)
            end_slice = spike + pd.Timedelta(noise_window)
            test_slice = dataseries[start_slice:end_slice].drop(spike)
            test_ratio_2 = np.abs(noise_func(test_slice) / test_slice.mean())
            # not a spike, we want to flag, if condition not satisfied:
            if test_ratio_2 > noise_thresh:
                spikes[spike] = False

        # not a spike, we want to flag, if condition not satisfied
        else:
            spikes[spike] = False

    spikes = spikes[spikes == True]

    flagger = flagger.setFlags(field, spikes.index, **kwargs)
    return data, flagger



#! /usr/bin/env python
# -*- coding: utf-8 -*-


class Fields:
    VARNAME = "varname"
    START = "start_date"
    END = "end_date"
    TESTS = "test*"
    PLOT = "plot"
    LINENUMBER = "line"


class Params:
    FLAG_GENERIC = "flagGeneric"
    GENERIC_ARGS = "func_arguments"
    FUNC = "func"



#! /usr/bin/env python
# -*- coding: utf-8 -*-


import re

from typing import Dict, List, Any, Union
from contextlib import contextmanager
from io import StringIO, TextIOWrapper

import numpy as np
import pandas as pd

from saqc.core.config import Fields as F
from saqc.core.evaluator import compileExpression
from saqc.flagger import BaseFlagger


ConfigList = List[Dict[str, Any]]


CONFIG_TYPES = {
    F.VARNAME: str,
    F.START: pd.to_datetime,
    F.END: pd.to_datetime,
    F.TESTS: str,
    F.PLOT: lambda v: str(v).lower() == "true",
    F.LINENUMBER: int,
}


def _raise(config_row, exc, msg, field=None):
    line_number = config_row[F.LINENUMBER]
    base_msg = f"configuration error in line {line_number}"
    if field:
        base_msg += f", column '{field}'"
    msg = base_msg + ":\n" + msg
    raise exc(msg)


@contextmanager
def _open(fname: str) -> Union[StringIO, TextIOWrapper]:
    if isinstance(fname, StringIO):
        yield fname
    else:
        f = open(fname)
        yield f
        f.close()


def _parseRow(row: str, sep: str, comment: str) -> List[str]:
    """
    remove in column comments, mainly needed to allow end line comments
    """
    return [c.split(comment)[0].strip() for c in row.split(sep)]


def _castRow(row: Dict[str, str]) -> Dict[str, Any]:
    """
    cast values to the data type given in 'types'
    """
    out = {}
    keys = pd.Index(row.keys())
    for k, func in CONFIG_TYPES.items():
        try:
            key = keys[keys.str.match(k)][0]
        except IndexError:
            continue
        value = row[key]
        # NOTE:
        # this check and the raise should be moved to checkConfig
        try:
            out[key] = func(value)
        except ValueError:
            _raise(row, ValueError, f"invalid value: '{value}'")
    return out


def _expandVarnameWildcards(config: ConfigList, data: pd.DataFrame) -> ConfigList:
    new = []
    for row in config:
        varname = row[F.VARNAME]
        if varname and varname not in data:
            expansion = data.columns[data.columns.str.match(varname)]
            if not len(expansion):
                expansion = [varname]
            for var in expansion:
                new.append({**row, F.VARNAME: var})
        else:
            new.append(row)
    return new


def readConfig(fname: str, data: pd.DataFrame, sep: str = ";", comment: str = "#") -> pd.DataFrame:

    defaults = {
        F.VARNAME: "",
        F.START: data.index.min(),
        F.END: data.index.max(),
        F.PLOT: False,
    }

    with _open(fname) as f:
        content = f.readlines()

    header: List = None
    config: ConfigList = []
    for i, line in enumerate(content):
        line = line.strip()
        if line.startswith(comment) or not line:
            continue
        row = _parseRow(line, sep, comment)
        if header is None:
            header = row
            continue
        values = dict(zip(header, row))
        values = {**defaults, **values, F.LINENUMBER: i + 1}
        config.append(_castRow(values))

    expanded = _expandVarnameWildcards(config, data)
    return pd.DataFrame(expanded)


def checkConfig(config_df: pd.DataFrame, data: pd.DataFrame, flagger: BaseFlagger, nodata: float) -> pd.DataFrame:
    for _, config_row in config_df.iterrows():

        var_name = config_row[F.VARNAME]
        if pd.isnull(config_row[F.VARNAME]) or not var_name:
            _raise(
                config_row, SyntaxError, f"non-optional column '{F.VARNAME}' is missing or empty",
            )

        test_fields = config_row.filter(regex=F.TESTS).dropna()
        if test_fields.empty:
            _raise(
                config_row, SyntaxError, f"at least one test needs to be given for variable",
            )

        for col, expr in test_fields.iteritems():
            if not expr:
                _raise(config_row, SyntaxError, f"field '{col}' may not be empty")
            try:
                compileExpression(expr, data, var_name, flagger, nodata)
            except (TypeError, NameError, SyntaxError) as exc:
                _raise(
                    config_row, type(exc), exc.args[0] + f" (failing statement: '{expr}')", col,
                )
    return config_df



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import numpy as np
import pandas as pd

from saqc.flagger import BaseFlagger

__plotvars = []

_colors = {
    "unflagged": "silver",
    "good": "seagreen",
    "bad": "firebrick",
    "suspicious": "gold",
}

_figsize = (10, 4)


def plotAllHook(data, flagger, plot_nans=False):
    if __plotvars:
        _plot(data, flagger, True, __plotvars, plot_nans=plot_nans)


def plotHook(
    data: pd.DataFrame,
    flagger_old: BaseFlagger,
    flagger_new: BaseFlagger,
    varname: str,
    flag_test: str,
    plot_nans: bool = False,
):

    # if data was harmonized, nans may occur in flags
    harm_nans = flagger_new.getFlags(varname).isna() | flagger_old.getFlags(varname).isna()

    # clean data from harmonisation nans
    if harm_nans.any():
        data = data.loc[~harm_nans, varname]

    if isinstance(data, pd.Series):
        data = data.to_frame()

    # clean flags from harmonisation nans
    try:
        flagger_old = flagger_old.getFlagger(varname, loc=data.index)
    except ValueError:
        # this might fail if someone want to plot the harmonisation itself,
        # but then we just plot the 'new' flags, ignoring the diff to the old ones
        mask = True
    else:
        flagger_new = flagger_new.getFlagger(varname, loc=data.index)

        # cannot use getFlags here, because if a flag was set (e.g. with force) the
        # flag may be the same, but any additional row (e.g. comment-field) would differ
        flags_old = flagger_old._flags[varname]
        flags_new = flagger_new._flags[varname]
        if len(flags_old) != len(flags_new):
            # NOTE:
            # we are just getting the result of an
            # harmonization, nothing to see here
            return

        mask = flags_old != flags_new

    __plotvars.append(varname)
    _plot(data, flagger_new, mask, varname, title=flag_test, plot_nans=plot_nans)


def _plot(
    data, flagger, flagmask, varname, interactive_backend=True, title="Data Plot", plot_nans=True,
):

    # todo: try catch warn (once) return
    # only import if plotting is requested by the user
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from pandas.plotting import register_matplotlib_converters

    # needed for datetime conversion
    register_matplotlib_converters()

    if not interactive_backend:
        # Import plot libs without interactivity, if not needed. This ensures that this can
        # produce an plot.png even if tkinter is not installed. E.g. if one want to run this
        # on machines without X-Server aka. graphic interface.
        mpl.use("Agg")
    else:
        mpl.use("TkAgg")

    if not isinstance(varname, (list, set)):
        varname = [varname]
    varname = set(varname)

    # filter out variables to which no data is associated (e.g. freshly assigned vars)
    tmp = []
    for var in varname:
        if var in data.columns:
            tmp.append(var)
        else:
            logging.warning(f"Cannot plot column '{var}', because it is not present in data.")
    if not tmp:
        return
    varnames = tmp

    plots = len(varnames)
    if plots > 1:
        fig, axes = plt.subplots(plots, 1, sharex=True, figsize=_figsize)
        axes[0].set_title(title)
        for i, v in enumerate(varnames):
            _plotByQualityFlag(data, v, flagger, flagmask, axes[i], plot_nans)
    else:
        fig, ax = plt.subplots(figsize=_figsize)
        plt.title(title)
        _plotByQualityFlag(data, varnames.pop(), flagger, flagmask, ax, plot_nans)

    # dummy plot for the label `missing` see _plotVline for more info
    if plot_nans:
        plt.plot([], [], ":", color="silver", label="missing data")

    plt.xlabel("time")
    plt.legend()

    if interactive_backend:
        plt.show()


def _plotByQualityFlag(data, varname, flagger, flagmask, ax, plot_nans):
    ax.set_ylabel(varname)

    if flagmask is True:
        flagmask = pd.Series(data=np.ones(len(data), dtype=bool), index=data.index)

    data = data[varname]
    if not plot_nans:
        data = data.dropna()
        flagmask = flagmask.loc[data.index]

    flagger = flagger.getFlagger(varname, loc=data.index)

    # base plot: show all(!) data
    ax.plot(
        data,
        # NOTE: no lines to flagged points
        # data.index, np.ma.array(data.values, mask=flagger.isFlagged(varname).values),
        "-",
        color="silver",
        label="data",
    )

    # ANY OLD FLAG
    # plot all(!) data that are already flagged in black
    flagged = flagger.isFlagged(varname, flag=flagger.GOOD, comparator=">=")
    oldflags = flagged & ~flagmask

    ax.plot(data[oldflags], ".", color="black", label="flagged by other test")
    if plot_nans:
        _plotNans(data[oldflags], "black", ax)

    # now we just want to show data that was flagged
    data = data.loc[flagmask[flagmask].index]
    flagger = flagger.getFlagger(varname, loc=data.index)

    if data.empty:
        return

    plots = [
        (flagger.UNFLAGGED, _colors["unflagged"]),
        (flagger.GOOD, _colors["good"]),
        (flagger.BAD, _colors["bad"]),
    ]

    for flag, color in plots:
        flagged = flagger.isFlagged(varname, flag=flag, comparator="==")
        if not data[flagged].empty:
            ax.plot(data[flagged], ".", color=color, label=f"flag: {flag}")
        if plot_nans:
            _plotNans(data[flagged], color, ax)

    # plot SUSPICIOS
    color = _colors["suspicious"]
    flagged = flagger.isFlagged(varname, flag=flagger.GOOD, comparator=">")
    flagged &= flagger.isFlagged(varname, flag=flagger.BAD, comparator="<")
    if not data[flagged].empty:
        ax.plot(
            data[flagged], ".", color=color, label=f"{flagger.GOOD} < flag < {flagger.BAD}",
        )
    if plot_nans:
        _plotNans(data[flagged], color, ax)


def _plotNans(y, color, ax):
    nans = y.isna()
    _plotVline(ax, y[nans].index, color=color)


def _plotVline(ax, points, color="blue"):
    # workaround for ax.vlines() as this work unexpected
    # normally this should work like so:
    #   ax.vlines(idx, *ylim, linestyles=':', color='silver', label="missing")
    for point in points:
        ax.axvline(point, color=color, linestyle=":")



#!/usr/bin/env python

from functools import partial
from inspect import signature, _VAR_KEYWORD


class Partial(partial):
    def __init__(self, func, *args, **kwargs):
        self._signature = signature(func)

    @property
    def signature(self):
        out = []
        for k, v in self._signature.parameters.items():
            if v.kind != _VAR_KEYWORD:
                out.append(k)
        return tuple(out)


# NOTE: will be filled by calls to register
FUNC_MAP = {}


def register(name):
    def outer(func):
        func = Partial(func, func_name=name)
        FUNC_MAP[name] = func

        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        return inner

    return outer



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd

from saqc.funcs.soil_moisture_tests import (
    flagSoilMoistureBySoilFrost,
    flagSoilMoistureByPrecipitationEvents,
    flagSoilMoistureConstant
)

from test.common import TESTFLAGGER, initData


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagSoilMoistureBySoilFrost(flagger):
    index = pd.date_range(
        start="2011-01-01 00:00:00", end="2011-01-01 03:00:00", freq="5min"
    )
    data = pd.DataFrame(
        {
            "soil_moisture": np.linspace(0, 1, index.size),
            "soil_temperature": np.linspace(1, -1, index.size),
        },
        index=index,
    )
    flagger = flagger.initFlags(data)
    data, flagger_result = flagSoilMoistureBySoilFrost(
        data, "soil_moisture", flagger, "soil_temperature"
    )
    flag_assertion = np.arange(19, 37)
    flag_result = flagger_result.getFlags("soil_moisture")
    assert (flag_result[flag_assertion] == flagger.BAD).all()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagSoilMoisturePrecipitationEvents(flagger):
    index = pd.date_range(
        start="2011-01-01 00:00:00", end="2011-01-04 00:00:00", freq="15min"
    )
    data = pd.DataFrame(
        {
            "soil_moisture": np.linspace(0, 1, index.size),
            "precipitation": np.linspace(1, 1, index.size),
        },
        index=index,
    )
    data["precipitation"]["2011-01-03"] = 0
    data["precipitation"]["2011-01-04"] = 0
    flagger = flagger.initFlags(data)
    data, flag_result = flagSoilMoistureByPrecipitationEvents(
        data, "soil_moisture", flagger, "precipitation"
    )
    flag_assertion = [288, 287]
    flag_result = flag_result.getFlags("soil_moisture")
    test_sum = (flag_result[flag_assertion] == flagger.BAD).sum()
    assert test_sum == len(flag_assertion)

@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagSoilMoistureConstants(flagger):

    data = initData(
        1, start_date="2011-01-01 00:00:00", end_date="2011-01-02 00:00:00", freq="5min"
    )
    data.iloc[5:25] = 0
    data.iloc[100:120] = data.max()[0]
    field = data.columns[0]
    flagger = flagger.initFlags(data)
    data, flagger = flagSoilMoistureConstant(
        data, field, flagger, window='1h', precipitation_window='1h')

    assert ~(flagger.isFlagged()[5:25]).all()[0]
    assert (flagger.isFlagged()[100:120]).all()[0]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from saqc.core.evaluator.evaluator import (
    compileExpression,
    evalExpression,
    compileTree,
    parseExpression,
    initLocalEnv,
    evalCode,
)

from saqc.core.evaluator.checker import DslChecker, ConfigChecker

from saqc.core.evaluator.transformer import DslTransformer, ConfigTransformer



from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="saqc",
    version="1.1.0",
    author="Bert Palm, David Schaefer, Peter Luenenschloss, Lennard Schmidt",
    author_email="david.schaefer@ufz.de",
    description="Data quality checking and processing tool/framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.ufz.de/rdm-software/saqc",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "pandas",
        "scipy",
        "scikit-learn==0.21.2",
        "numba",
        "matplotlib",
        "click",
        "pyarrow",
        "python-intervals",
    ],
    license="GPLv3",
    entry_points={"console_scripts": ["saqc=saqc.__main__:main"],},
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd

from saqc.core.evaluator import evalExpression
from saqc.funcs.functions import (
    flagRange,
    flagSesonalRange,
    forceFlags,
    clearFlags,
    flagIsolated,
)
from saqc.flagger.dmpflagger import DmpFlagger
from test.common import initData, TESTFLAGGER


@pytest.fixture
def data():
    return initData(cols=1, start_date="2016-01-01", end_date="2018-12-31", freq="1D")


@pytest.fixture
def field(data):
    return data.columns[0]


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagAfter(data, field, flagger):

    flagger = flagger.initFlags(data)

    min = data.iloc[int(len(data) * 0.3), 0]
    max = data.iloc[int(len(data) * 0.6), 0]
    _, flagger_range = flagRange(data, field, flagger, min, max)
    flagged_range = flagger_range.isFlagged(field, loc=flagger_range.isFlagged(field))

    tests = [
        (f"flagWindowAfterFlag(window='3D', func=range(min={min}, max={max}))", "3D"),
        (f"flagNextAfterFlag(n=4, func=range(min={min}, max={max}))", 4),
    ]

    for expr, window in tests:
        _, flagger_range_repeated = evalExpression(expr, data, field, flagger)

        check = flagged_range.rolling(window=window).apply(
            lambda df: (flagger_range_repeated.isFlagged(field, loc=df.index).all()),
            raw=False,
        )
        assert check.all()


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_range(data, field, flagger):
    min, max = 10, 90
    flagger = flagger.initFlags(data)
    data, flagger = flagRange(data, field, flagger, min=min, max=max)
    flagged = flagger.isFlagged(field)
    expected = (data[field] < min) | (data[field] > max)
    assert np.all(flagged == expected)


# @pytest.mark.parametrize('flagger', TESTFLAGGER)
# def test_missing(data, field, flagger):
#     pass


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagSesonalRange(data, field, flagger):
    # prepare
    data.loc[::2] = 0
    data.loc[1::2] = 50
    nyears = len(data.index.year.unique())

    tests = [
        (
            {
                "min": 1,
                "max": 100,
                "startmonth": 7,
                "startday": 1,
                "endmonth": 8,
                "endday": 31,
            },
            31 * 2 * nyears // 2,
        ),
        (
            {
                "min": 1,
                "max": 100,
                "startmonth": 12,
                "startday": 16,
                "endmonth": 1,
                "endday": 15,
            },
            31 * nyears // 2 + 1,
        ),
    ]

    for test, expected in tests:
        flagger = flagger.initFlags(data)
        data, flagger = flagSesonalRange(data, field, flagger, **test)
        flagged = flagger.isFlagged(field)
        assert flagged.sum() == expected


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_clearFlags(data, field, flagger):
    flagger = flagger.initFlags(data)
    flags_orig = flagger.getFlags()
    flags_set = flagger.setFlags(field, flag=flagger.BAD).getFlags()
    _, flagger = clearFlags(data, field, flagger)
    flags_cleared = flagger.getFlags()
    assert np.all(flags_orig != flags_set)
    assert np.all(flags_orig == flags_cleared)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_forceFlags(data, flagger):
    flagger = flagger.initFlags(data)
    field, *_ = data.columns
    flags_orig = flagger.setFlags(field).getFlags(field)
    _, flagger = forceFlags(data, field, flagger, flag=flagger.GOOD)
    flags_forced = flagger.getFlags(field)
    assert np.all(flags_orig != flags_forced)


@pytest.mark.parametrize("flagger", TESTFLAGGER)
def test_flagIsolated(data, flagger):
    field = data.columns[0]
    data.iloc[1:3, 0] = np.nan
    data.iloc[4:5, 0] = np.nan
    data.iloc[11:13, 0] = np.nan
    data.iloc[15:17, 0] = np.nan
    flagger = flagger.initFlags(data)
    flagger = flagger.setFlags(field, iloc=slice(5, 6))

    _, flagger_result = flagIsolated(data, field, flagger, group_window="1D", gap_window="2.1D")

    assert flagger_result.isFlagged(field)[slice(3, 6, 2)].all()

    flagger = flagger.setFlags(
        field, iloc=slice(3, 4), flag=flagger.UNFLAGGED, force=True
    )
    data, flagger_result = flagIsolated(
        data,
        field,
        flagger_result,
        group_window="2D",
        gap_window="2.1D",
        continuation_range="1.1D",
    )
    assert flagger_result.isFlagged(field)[[3, 5, 13, 14]].all()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
import intervals

from saqc.flagger.baseflagger import BaseFlagger


class ContinuousFlagger(BaseFlagger):
    def __init__(self, min_=0.0, max_=1.0, unflagged=-1.0):
        assert unflagged < 0 <= min_ < max_
        super().__init__(dtype=np.float64)
        self._interval = intervals.closed(min_, max_)
        self._unflagged_flag = unflagged
        self.signature = ("flag", "factor", "modify")

    def setFlags(self, field, loc=None, iloc=None, flag=None, force=False, factor=1, modify=False, **kwargs):
        # NOTE: incomplete, as the option to
        #       update flags is temporarily gone
        return super().setFlags(field=field, loc=loc, iloc=iloc, flag=self._checkFlag(flag), force=force, **kwargs)

    # NOTE:
    # we should probably override _assureDtype here

    def _isDtype(self, flag):
        if isinstance(flag, pd.Series):
            # NOTE: it should be made sure, that all
            #       values fall into the interval
            return flag.dtype == self.dtype
        return flag in self._interval or flag == self.UNFLAGGED

    @property
    def UNFLAGGED(self):
        return self._unflagged_flag

    @property
    def GOOD(self):
        return self._interval.lower

    @property
    def BAD(self):
        return self._interval.upper

    def isSUSPICIOUS(self, flag):
        return flag in intervals.open(self.GOOD, self.BAD)
